"""Tests for qmm.extensions.life module."""

import pytest
import sympy as sp

from qmm.core.helper import get_nodes
from qmm.extensions.life import (
    birth_matrix,
    death_matrix,
    life_expectancy_change,
    net_life_expectancy_change,
    absolute_life_expectancy_change,
    weighted_predictions_life_expectancy,
)

# =============================================================================
# birth_matrix
# =============================================================================

def test_birth_matrix_form_signed_snowshoe(snowshoe):
    result = birth_matrix(snowshoe, form='signed')
    expected = sp.Matrix([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0]])
    assert result == expected


def test_birth_matrix_form_symbolic_snowshoe(snowshoe):
    result = birth_matrix(snowshoe, form='symbolic')
    expected = sp.Matrix([
        [                 0,                  0, 0],
        [sp.Symbol('a_C,R'),                  0, 0],
        [                 0, sp.Symbol('a_P,C'), 0]])
    assert result == expected


def test_birth_matrix_form_signed_perturb_first_state_snowshoe(snowshoe):
    nodes = get_nodes(snowshoe, 'state')
    result = birth_matrix(snowshoe, form='signed', perturb=nodes[0])
    expected = sp.Matrix([0, 1, 0])
    assert result == expected


def test_birth_matrix_form_symbolic_perturb_first_state_snowshoe(snowshoe):
    nodes = get_nodes(snowshoe, 'state')
    result = birth_matrix(snowshoe, form='symbolic', perturb=nodes[0])
    expected = sp.Matrix([0, sp.Symbol('a_C,R'), 0])
    assert result == expected


def test_birth_matrix_form_signed_chain(chain):
    result = birth_matrix(chain, form='signed')
    expected = sp.Matrix([
        [0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0],
        [0, 1, 0, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 0, 1, 0]])
    assert result == expected


def test_birth_matrix_form_signed_mesocosm(mesocosm):
    result = birth_matrix(mesocosm, form='signed')
    expected = sp.Matrix([
        [0, 0, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 1, 0]])
    assert result == expected

# =============================================================================
# death_matrix
# =============================================================================

def test_death_matrix_form_signed_snowshoe(snowshoe):
    result = death_matrix(snowshoe, form='signed')
    expected = sp.Matrix([
        [1, 1, 0],
        [0, 0, 1],
        [0, 0, 1]])
    assert result == expected


def test_death_matrix_form_symbolic_snowshoe(snowshoe):
    result = death_matrix(snowshoe, form='symbolic')
    expected = sp.Matrix([
        [sp.Symbol('a_R,R'), sp.Symbol('a_R,C'),                  0],
        [                 0,                  0, sp.Symbol('a_C,P')],
        [                 0,                  0, sp.Symbol('a_P,P')]])
    assert result == expected


def test_death_matrix_form_signed_perturb_R_snowshoe(snowshoe):
    result = death_matrix(snowshoe, form='signed', perturb='R')
    expected = sp.Matrix([1, 0, 0])
    assert result == expected


def test_death_matrix_form_symbolic_perturb_first_state_snowshoe(snowshoe):
    nodes = get_nodes(snowshoe, 'state')
    result = death_matrix(snowshoe, form='symbolic', perturb=nodes[0])
    expected = sp.Matrix([sp.Symbol('a_R,R'), 0, 0])
    assert result == expected


def test_death_matrix_form_signed_chain(chain):
    result = death_matrix(chain, form='signed')
    expected = sp.Matrix([
        [1, 1, 0, 0, 0],
        [0, 1, 1, 0, 0],
        [0, 0, 1, 1, 0],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 0, 1]])
    assert result == expected


def test_death_matrix_form_signed_mesocosm(mesocosm):
    result = death_matrix(mesocosm, form='signed')
    expected = sp.Matrix([
        [1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 1]])
    assert result == expected

# =============================================================================
# life_expectancy_change
# =============================================================================

def test_life_expectancy_change_form_signed_birth_snowshoe(snowshoe):
    result = life_expectancy_change(snowshoe, form='signed', type='birth')
    expected = sp.Matrix([
        [-2,  0,  0],
        [-1, -1, -1],
        [-1, -1, -1]])
    assert result == expected


def test_life_expectancy_change_form_signed_death_snowshoe(snowshoe):
    result = life_expectancy_change(snowshoe, form='signed', type='death')
    expected = sp.Matrix([
        [ 0,  0,  0],
        [-1,  1, -1],
        [-1, -1,  1]])
    assert result == expected


def test_life_expectancy_change_form_signed_birth_perturb_first_state_snowshoe(snowshoe):
    nodes = get_nodes(snowshoe, 'state')
    result = life_expectancy_change(snowshoe, form='signed', type='birth', perturb=nodes[0])
    expected = sp.Matrix([-2, -1, -1])
    assert result == expected


def test_life_expectancy_change_invalid_perturb_node(snowshoe):
    with pytest.raises(ValueError, match="Perturbation node must be one of"):
        life_expectancy_change(snowshoe, form='signed', type='birth', perturb='InvalidNode')


def test_life_expectancy_change_form_signed_birth_chain(chain):
    result = life_expectancy_change(chain, form='signed', type='birth')
    expected = sp.Matrix([
        [-8,  0,  0,  0,  0],
        [-5, -5, -2,  1, -1],
        [-3, -3, -6, -1,  1],
        [-2, -2, -4, -6, -2],
        [-1, -1, -2, -3, -5]])
    assert result == expected


def test_life_expectancy_change_form_symbolic_birth_snowshoe(snowshoe):
    result = life_expectancy_change(snowshoe, form='symbolic', type='birth')
    expected = sp.Matrix([
        [-sp.Symbol('a_C,P')*sp.Symbol('a_P,C')*sp.Symbol('a_R,R') - sp.Symbol('a_C,R')*sp.Symbol('a_P,P')*sp.Symbol('a_R,C'),                                                         0,                                                         0],
        [                                                           -sp.Symbol('a_C,P')*sp.Symbol('a_C,R')*sp.Symbol('a_P,C'), -sp.Symbol('a_C,P')*sp.Symbol('a_P,C')*sp.Symbol('a_R,R'), -sp.Symbol('a_C,P')*sp.Symbol('a_C,R')*sp.Symbol('a_R,C')],
        [                                                           -sp.Symbol('a_C,R')*sp.Symbol('a_P,C')*sp.Symbol('a_P,P'), -sp.Symbol('a_P,C')*sp.Symbol('a_P,P')*sp.Symbol('a_R,R'), -sp.Symbol('a_C,R')*sp.Symbol('a_P,P')*sp.Symbol('a_R,C')]])
    assert result == expected


def test_life_expectancy_change_form_symbolic_death_snowshoe(snowshoe):
    result = life_expectancy_change(snowshoe, form='symbolic', type='death')
    expected = sp.Matrix([
        [                                                        0,                                                         0,                                                         0],
        [-sp.Symbol('a_C,P')*sp.Symbol('a_C,R')*sp.Symbol('a_P,C'),  sp.Symbol('a_C,R')*sp.Symbol('a_P,P')*sp.Symbol('a_R,C'), -sp.Symbol('a_C,P')*sp.Symbol('a_C,R')*sp.Symbol('a_R,C')],
        [-sp.Symbol('a_C,R')*sp.Symbol('a_P,C')*sp.Symbol('a_P,P'), -sp.Symbol('a_P,C')*sp.Symbol('a_P,P')*sp.Symbol('a_R,R'),  sp.Symbol('a_C,P')*sp.Symbol('a_P,C')*sp.Symbol('a_R,R')]])
    assert result == expected

# =============================================================================
# net_life_expectancy_change
# =============================================================================

def test_net_life_expectancy_change_birth_snowshoe(snowshoe):
    result = net_life_expectancy_change(snowshoe, type='birth')
    expected = sp.Matrix([
        [-2,  0,  0],
        [-1, -1, -1],
        [-1, -1, -1]])
    assert result == expected


def test_net_life_expectancy_change_death_snowshoe(snowshoe):
    result = net_life_expectancy_change(snowshoe, type='death')
    expected = sp.Matrix([
        [ 0,  0,  0],
        [-1,  1, -1],
        [-1, -1,  1]])
    assert result == expected


def test_net_life_expectancy_change_birth_chain(chain):
    result = net_life_expectancy_change(chain, type='birth')
    expected = sp.Matrix([
        [-8,  0,  0,  0,  0],
        [-5, -5, -2,  1, -1],
        [-3, -3, -6, -1,  1],
        [-2, -2, -4, -6, -2],
        [-1, -1, -2, -3, -5]])
    assert result == expected


def test_net_life_expectancy_change_birth_mesocosm(mesocosm):
    result = net_life_expectancy_change(mesocosm, type='birth')
    expected = sp.Matrix([
        [-2,  0,  0,  0, 0,  0,  0, 0],
        [-1, -1, -1,  1, 0, -1,  1, 0],
        [-1,  1, -3,  1, 0, -1,  1, 0],
        [-1,  1, -1, -1, 0, -1,  1, 0],
        [ 1, -3,  5, -1, 0,  1, -5, 2],
        [ 0, -2,  2,  0, 0,  0, -2, 0],
        [ 0, -2,  2,  0, 0,  0, -2, 0],
        [ 0, -2,  2,  0, 0,  0, -2, 0]])
    assert result == expected


def test_net_life_expectancy_change_death_mesocosm(mesocosm):
    result = net_life_expectancy_change(mesocosm, type='death')
    expected = sp.Matrix([
        [ 0,  0,  0,  0, 0,  0,  0, 0],
        [-1,  1, -1,  1, 0, -1,  1, 0],
        [-1,  1, -1,  1, 0, -1,  1, 0],
        [-1,  1, -1,  1, 0, -1,  1, 0],
        [ 1, -3,  5, -1, 2,  1, -5, 2],
        [ 0, -2,  2,  0, 0,  2, -2, 0],
        [ 0, -2,  2,  0, 0,  0,  0, 0],
        [ 0, -2,  2,  0, 0,  0, -2, 2]])
    assert result == expected


# =============================================================================
# absolute_life_expectancy_change
# =============================================================================

def test_absolute_life_expectancy_change_birth_snowshoe(snowshoe):
    result = absolute_life_expectancy_change(snowshoe, type='birth')
    expected = sp.Matrix([
        [2, 0, 0],
        [1, 1, 1],
        [1, 1, 1]])
    assert result == expected


def test_absolute_life_expectancy_change_death_snowshoe(snowshoe):
    result = absolute_life_expectancy_change(snowshoe, type='death')
    expected = sp.Matrix([
        [0, 0, 0],
        [1, 1, 1],
        [1, 1, 1]])
    assert result == expected


def test_absolute_life_expectancy_change_death_chain(chain):
    result = absolute_life_expectancy_change(chain, type='death')
    expected = sp.Matrix([
        [0, 0, 0, 0, 0],
        [5, 3, 2, 1, 1],
        [3, 3, 2, 1, 1],
        [2, 2, 4, 2, 2],
        [1, 1, 2, 3, 3]])
    assert result == expected


def test_absolute_life_expectancy_change_birth_mesocosm(mesocosm):
    result = absolute_life_expectancy_change(mesocosm, type='birth')
    expected = sp.Matrix([
        [18,  0, 0,  0,  0, 0, 0, 0],
        [ 1, 11, 9,  1,  2, 1, 5, 2],
        [ 1,  7, 9,  1,  2, 1, 5, 2],
        [ 1,  7, 9, 17,  2, 1, 5, 2],
        [ 7,  7, 9,  7, 14, 7, 7, 4],
        [ 2,  4, 4,  2,  4, 2, 8, 4],
        [ 2,  4, 4,  2,  4, 2, 8, 4],
        [ 2,  4, 4,  2,  4, 2, 8, 4]])
    assert result == expected


def test_absolute_life_expectancy_change_death_mesocosm(mesocosm):
    result = absolute_life_expectancy_change(mesocosm, type='death')
    expected = sp.Matrix([
        [0, 0, 0, 0, 0,  0,  0,  0],
        [1, 7, 9, 1, 2,  1,  5,  2],
        [1, 7, 9, 1, 2,  1,  5,  2],
        [1, 7, 9, 1, 2,  1,  5,  2],
        [7, 7, 9, 7, 4,  7,  7,  4],
        [2, 4, 4, 2, 4, 16,  8,  4],
        [2, 4, 4, 2, 4,  2, 10,  4],
        [2, 4, 4, 2, 4,  2,  8, 14]])
    assert result == expected


# =============================================================================
# weighted_predictions_life_expectancy
# =============================================================================

def test_weighted_predictions_life_expectancy_birth_snowshoe(snowshoe):
    result = weighted_predictions_life_expectancy(snowshoe, type='birth')
    expected = sp.Matrix([
        [-1, sp.nan, sp.nan],
        [-1,     -1,     -1],
        [-1,     -1,     -1]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_as_nan_false_snowshoe(snowshoe):
    result = weighted_predictions_life_expectancy(snowshoe, type='birth', as_nan=False)
    expected = sp.Matrix([
        [-1,  1,  1],
        [-1, -1, -1],
        [-1, -1, -1]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_as_abs_true_snowshoe(snowshoe):
    result = weighted_predictions_life_expectancy(snowshoe, type='birth', as_abs=True)
    expected = sp.Matrix([
        [1, sp.nan, sp.nan],
        [1,      1,      1],
        [1,      1,      1]])
    assert result == expected


def test_weighted_predictions_life_expectancy_death_snowshoe(snowshoe):
    result = weighted_predictions_life_expectancy(snowshoe, type='death')
    expected = sp.Matrix([
        [sp.nan, sp.nan, sp.nan],
        [    -1,      1,     -1],
        [    -1,     -1,      1]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_chain(chain):
    result = weighted_predictions_life_expectancy(chain, type='birth')
    expected = sp.Matrix([
        [-1, sp.nan, sp.nan, sp.nan, sp.nan],
        [-1,     -1,     -1,      1,     -1],
        [-1,     -1,     -1,     -1,      1],
        [-1,     -1,     -1,     -1,     -1],
        [-1,     -1,     -1,     -1,     -1]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='birth')
    expected = sp.Matrix([
        [sp.Rational(-1, 9),              sp.nan,             sp.nan,              sp.nan, sp.nan,            sp.nan,             sp.nan,            sp.nan],
        [                -1, sp.Rational(-1, 11), sp.Rational(-1, 9),                   1,      0,                -1,  sp.Rational(1, 5),                 0],
        [                -1,   sp.Rational(1, 7), sp.Rational(-1, 3),                   1,      0,                -1,  sp.Rational(1, 5),                 0],
        [                -1,   sp.Rational(1, 7), sp.Rational(-1, 9), sp.Rational(-1, 17),      0,                -1,  sp.Rational(1, 5),                 0],
        [ sp.Rational(1, 7),  sp.Rational(-3, 7),  sp.Rational(5, 9),  sp.Rational(-1, 7),      0, sp.Rational(1, 7), sp.Rational(-5, 7), sp.Rational(1, 2)],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0,      0,                 0, sp.Rational(-1, 4),                 0],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0,      0,                 0, sp.Rational(-1, 4),                 0],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0,      0,                 0, sp.Rational(-1, 4),                 0]])
    assert result == expected


def test_weighted_predictions_life_expectancy_death_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='death')
    expected = sp.Matrix([
        [           sp.nan,             sp.nan,             sp.nan,             sp.nan,            sp.nan,            sp.nan,             sp.nan,            sp.nan],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [sp.Rational(1, 7), sp.Rational(-3, 7),  sp.Rational(5, 9), sp.Rational(-1, 7), sp.Rational(1, 2), sp.Rational(1, 7), sp.Rational(-5, 7), sp.Rational(1, 2)],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0, sp.Rational(1, 8), sp.Rational(-1, 4),                 0],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0,                 0,                  0,                 0],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0,                 0, sp.Rational(-1, 4), sp.Rational(1, 7)]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_as_nan_false_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='birth', as_nan=False)
    expected = sp.Matrix([
        [sp.Rational(-1, 9),                   1,                  1,                   1, 1,                 1,                  1,                 1],
        [                -1, sp.Rational(-1, 11), sp.Rational(-1, 9),                   1, 0,                -1,  sp.Rational(1, 5),                 0],
        [                -1,   sp.Rational(1, 7), sp.Rational(-1, 3),                   1, 0,                -1,  sp.Rational(1, 5),                 0],
        [                -1,   sp.Rational(1, 7), sp.Rational(-1, 9), sp.Rational(-1, 17), 0,                -1,  sp.Rational(1, 5),                 0],
        [ sp.Rational(1, 7),  sp.Rational(-3, 7),  sp.Rational(5, 9),  sp.Rational(-1, 7), 0, sp.Rational(1, 7), sp.Rational(-5, 7), sp.Rational(1, 2)],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0, 0,                 0, sp.Rational(-1, 4),                 0],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0, 0,                 0, sp.Rational(-1, 4),                 0],
        [                 0,  sp.Rational(-1, 2),  sp.Rational(1, 2),                   0, 0,                 0, sp.Rational(-1, 4),                 0]])
    assert result == expected


def test_weighted_predictions_life_expectancy_birth_as_abs_true_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='birth', as_abs=True)
    expected = sp.Matrix([
        [sp.Rational(1, 9),             sp.nan,            sp.nan,             sp.nan, sp.nan,            sp.nan,            sp.nan,            sp.nan],
        [                1, sp.Rational(1, 11), sp.Rational(1, 9),                  1,      0,                 1, sp.Rational(1, 5),                 0],
        [                1,  sp.Rational(1, 7), sp.Rational(1, 3),                  1,      0,                 1, sp.Rational(1, 5),                 0],
        [                1,  sp.Rational(1, 7), sp.Rational(1, 9), sp.Rational(1, 17),      0,                 1, sp.Rational(1, 5),                 0],
        [sp.Rational(1, 7),  sp.Rational(3, 7), sp.Rational(5, 9),  sp.Rational(1, 7),      0, sp.Rational(1, 7), sp.Rational(5, 7), sp.Rational(1, 2)],
        [                0,  sp.Rational(1, 2), sp.Rational(1, 2),                  0,      0,                 0, sp.Rational(1, 4),                 0],
        [                0,  sp.Rational(1, 2), sp.Rational(1, 2),                  0,      0,                 0, sp.Rational(1, 4),                 0],
        [                0,  sp.Rational(1, 2), sp.Rational(1, 2),                  0,      0,                 0, sp.Rational(1, 4),                 0]])
    assert result == expected


def test_weighted_predictions_life_expectancy_death_as_nan_false_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='death', as_nan=False)
    expected = sp.Matrix([
        [                1,                  1,                  1,                  1,                 1,                 1,                  1,                 1],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [               -1,  sp.Rational(1, 7), sp.Rational(-1, 9),                  1,                 0,                -1,  sp.Rational(1, 5),                 0],
        [sp.Rational(1, 7), sp.Rational(-3, 7),  sp.Rational(5, 9), sp.Rational(-1, 7), sp.Rational(1, 2), sp.Rational(1, 7), sp.Rational(-5, 7), sp.Rational(1, 2)],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0, sp.Rational(1, 8), sp.Rational(-1, 4),                 0],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0,                 0,                  0,                 0],
        [                0, sp.Rational(-1, 2),  sp.Rational(1, 2),                  0,                 0,                 0, sp.Rational(-1, 4), sp.Rational(1, 7)]])
    assert result == expected


def test_weighted_predictions_life_expectancy_death_as_abs_true_mesocosm(mesocosm):
    result = weighted_predictions_life_expectancy(mesocosm, type='death', as_abs=True)
    expected = sp.Matrix([
        [           sp.nan,            sp.nan,            sp.nan,            sp.nan,            sp.nan,            sp.nan,            sp.nan,            sp.nan],
        [                1, sp.Rational(1, 7), sp.Rational(1, 9),                 1,                 0,                 1, sp.Rational(1, 5),                 0],
        [                1, sp.Rational(1, 7), sp.Rational(1, 9),                 1,                 0,                 1, sp.Rational(1, 5),                 0],
        [                1, sp.Rational(1, 7), sp.Rational(1, 9),                 1,                 0,                 1, sp.Rational(1, 5),                 0],
        [sp.Rational(1, 7), sp.Rational(3, 7), sp.Rational(5, 9), sp.Rational(1, 7), sp.Rational(1, 2), sp.Rational(1, 7), sp.Rational(5, 7), sp.Rational(1, 2)],
        [                0, sp.Rational(1, 2), sp.Rational(1, 2),                 0,                 0, sp.Rational(1, 8), sp.Rational(1, 4),                 0],
        [                0, sp.Rational(1, 2), sp.Rational(1, 2),                 0,                 0,                 0,                 0,                 0],
        [                0, sp.Rational(1, 2), sp.Rational(1, 2),                 0,                 0,                 0, sp.Rational(1, 4), sp.Rational(1, 7)]])
    assert result == expected


def test_weighted_predictions_life_expectancy_invalid_type_snowshoe(snowshoe):
    with pytest.raises(ValueError) as exc_info:
        weighted_predictions_life_expectancy(snowshoe, type='invalid')
    result = str(exc_info.value)
    expected = "type must be either 'birth' or 'death'"
    assert result == expected


# =============================================================================
# Additional coverage tests
# =============================================================================

def test_birth_matrix_invalid_perturb_node(snowshoe):
    with pytest.raises(ValueError, match="Perturbation node must be one of"):
        birth_matrix(snowshoe, perturb="Invalid")


def test_death_matrix_invalid_perturb_node(snowshoe):
    with pytest.raises(ValueError, match="Perturbation node must be one of"):
        death_matrix(snowshoe, perturb="Invalid")


def test_life_expectancy_change_invalid_type(snowshoe):
    with pytest.raises(ValueError, match="type must be either 'birth' or 'death'"):
        life_expectancy_change(snowshoe, type="invalid")


def test_weighted_predictions_life_expectancy_invalid_type_coverage(snowshoe):
    with pytest.raises(ValueError, match="type must be either 'birth' or 'death'"):
        weighted_predictions_life_expectancy(snowshoe, type="invalid")




"""Tests for qmm.extensions.effects module."""

import pytest
import sympy as sp
import numpy as np
import pandas as pd

from qmm.core.helper import get_nodes, _check_acyclic_inputs, _check_acyclic_outputs
from qmm.extensions.effects import define_input_output
from qmm.extensions.effects import (
    cumulative_effects,
    net_effects,
    absolute_effects,
    weighted_effects,
    sign_determinacy_effects,
    get_simulations,
    simulation_effects,
    simulations_table,
    direct_effects,
    table_of_direct_effects,
    table_of_effects,
)
from qmm.core.press import (
    adjoint_matrix,
    numerical_simulations,
    absolute_feedback_matrix,
    weighted_predictions_matrix,
    sign_determinacy_matrix,
)

# =============================================================================
# define_input_output
# =============================================================================

def test_define_input_output_categories_snowshoe_io(snowshoe_io):
    categorized = define_input_output(snowshoe_io)
    result = {node: data['category'] for node, data in categorized.nodes(data=True)}
    expected = {
        'R': 'state',
        'C': 'state',
        'P': 'state',
        'Inp1': 'input',
        'Inp2': 'input',
        'Out1': 'output',
        'Out2': 'output',
    }
    assert result == expected


def test_define_input_output_remove_disconnected_true_disconnected_graph(disconnected_graph):
    result = sorted(define_input_output(disconnected_graph, remove_disconnected=True).nodes())
    expected = ['A', 'B']
    assert result == expected


def test_define_input_output_remove_disconnected_false_disconnected_graph(disconnected_graph):
    result = sorted(define_input_output(disconnected_graph, remove_disconnected=False).nodes())
    expected = ['A', 'B', 'C']
    assert result == expected


def test_define_input_output_invalid_input_type_no_fixture():
    with pytest.raises(TypeError) as exc_info:
        define_input_output("not a graph")
    result = str(exc_info.value)
    expected = "Input must be a networkx.DiGraph."
    assert result == expected


def test_define_input_output_rejects_precategorized_cyclic_inputs(cyclic_inputs_graph):
    with pytest.raises(ValueError) as exc_info:
        _check_acyclic_inputs(cyclic_inputs_graph)
    result = str(exc_info.value)
    expected = "Input subgraph contains cycles - input nodes must be acyclic"
    assert result == expected


def test_define_input_output_rejects_precategorized_cyclic_outputs(cyclic_outputs_graph):
    with pytest.raises(ValueError) as exc_info:
        _check_acyclic_outputs(cyclic_outputs_graph)
    result = str(exc_info.value)
    expected = "Output subgraph contains cycles - output nodes must be acyclic"
    assert result == expected


def test_define_input_output_rejects_feedthrough(snowshoe_io_with_direct_edge):
    with pytest.raises(ValueError) as exc_info:
        define_input_output(snowshoe_io_with_direct_edge)
    assert "Direct input to output edge" in str(exc_info.value)

# =============================================================================
# cumulative_effects
# =============================================================================

def test_cumulative_effects_form_signed_snowshoe_io(snowshoe_io):
    result = cumulative_effects(snowshoe_io, form='signed')
    expected = sp.Matrix([
        [1, -1,  1, 2, -1],
        [1,  1, -1, 0,  1],
        [1,  1,  1, 0, -1],
        [0,  0,  2, 0, -2],
        [1,  1, -1, 0,  1]])
    assert result == expected


def test_cumulative_effects_form_binary_snowshoe_io(snowshoe_io):
    result = cumulative_effects(snowshoe_io, form='binary')
    expected = sp.Matrix([
        [1, 1, 1, 2, 1],
        [1, 1, 1, 2, 1],
        [1, 1, 1, 2, 1],
        [2, 2, 2, 4, 2],
        [1, 1, 1, 2, 1]])
    assert result == expected


def test_cumulative_effects_form_symbolic_snowshoe_io(snowshoe_io):
    result = cumulative_effects(snowshoe_io, form='symbolic')
    a_RR = sp.Symbol('a_R,R')
    a_RC = sp.Symbol('a_R,C')
    a_CR = sp.Symbol('a_C,R')
    a_CP = sp.Symbol('a_C,P')
    a_PC = sp.Symbol('a_P,C')
    a_PP = sp.Symbol('a_P,P')
    b_R_Inp1 = sp.Symbol('b_R,Inp1')
    b_C_Inp1 = sp.Symbol('b_C,Inp1')
    b_P_Inp2 = sp.Symbol('b_P,Inp2')
    c_Out1_C = sp.Symbol('c_Out1,C')
    c_Out1_P = sp.Symbol('c_Out1,P')
    c_Out2_C = sp.Symbol('c_Out2,C')
    expected = sp.Matrix([
        [                              a_CP*a_PC,                              -a_PP*a_RC,                               a_CP*a_RC,                                                                               a_CP*a_PC*b_R_Inp1 + a_PP*a_RC*b_C_Inp1,                                        -a_CP*a_RC*b_P_Inp2],
        [                              a_CR*a_PP,                               a_PP*a_RR,                              -a_CP*a_RR,                                                                               a_CR*a_PP*b_R_Inp1 - a_PP*a_RR*b_C_Inp1,                                         a_CP*a_RR*b_P_Inp2],
        [                              a_CR*a_PC,                               a_PC*a_RR,                               a_CR*a_RC,                                                                               a_CR*a_PC*b_R_Inp1 - a_PC*a_RR*b_C_Inp1,                                        -a_CR*a_RC*b_P_Inp2],
        [a_CR*a_PC*c_Out1_P - a_CR*a_PP*c_Out1_C, a_PC*a_RR*c_Out1_P - a_PP*a_RR*c_Out1_C, a_CP*a_RR*c_Out1_C + a_CR*a_RC*c_Out1_P, a_CR*a_PC*b_R_Inp1*c_Out1_P - a_CR*a_PP*b_R_Inp1*c_Out1_C - a_PC*a_RR*b_C_Inp1*c_Out1_P + a_PP*a_RR*b_C_Inp1*c_Out1_C, -a_CP*a_RR*b_P_Inp2*c_Out1_C - a_CR*a_RC*b_P_Inp2*c_Out1_P],
        [                     a_CR*a_PP*c_Out2_C,                      a_PP*a_RR*c_Out2_C,                     -a_CP*a_RR*c_Out2_C,                                                             a_CR*a_PP*b_R_Inp1*c_Out2_C - a_PP*a_RR*b_C_Inp1*c_Out2_C,                                a_CP*a_RR*b_P_Inp2*c_Out2_C]])
    assert result == expected


def test_cumulative_effects_form_signed_snowshoe(snowshoe):
    G = define_input_output(snowshoe)
    result = cumulative_effects(G, form='signed')
    expected = sp.Matrix([
        [1, -1,  1],
        [1,  1, -1],
        [1,  1,  1]])
    assert result == expected


def test_cumulative_effects_symbolic_matches_adjoint_snowshoe_snowshoe_io(snowshoe, snowshoe_io):
    result = cumulative_effects(snowshoe_io, form='symbolic')[:3, :3]
    expected = adjoint_matrix(snowshoe, form='symbolic')
    assert result == expected


def test_cumulative_effects_invalid_form_snowshoe_io(snowshoe_io):
    with pytest.raises(ValueError) as exc_info:
        cumulative_effects(snowshoe_io, form='invalid')
    result = str(exc_info.value)
    expected = "Invalid form. Choose 'symbolic', 'signed', 'binary'."
    assert result == expected

# =============================================================================
# absolute_effects
# =============================================================================

def test_absolute_effects_default_snowshoe_io(snowshoe_io):
    result = absolute_effects(snowshoe_io)
    expected = sp.Matrix([
        [1, 1, 1, 2, 1],
        [1, 1, 1, 2, 1],
        [1, 1, 1, 2, 1],
        [2, 2, 2, 4, 2],
        [1, 1, 1, 2, 1]])
    assert result == expected


def test_absolute_effects_match_binary_cumulative_snowshoe_io(snowshoe_io):
    result = absolute_effects(snowshoe_io)
    expected = cumulative_effects(snowshoe_io, form='binary')
    assert result == expected


def test_absolute_effects_vs_absolute_feedback_snowshoe_snowshoe_io(snowshoe, snowshoe_io):
    result = absolute_effects(snowshoe_io)[:3, :3]
    expected = absolute_feedback_matrix(snowshoe)
    assert result == expected

# =============================================================================
# weighted_effects
# =============================================================================

def test_weighted_effects_snowshoe_io(snowshoe_io):
    result = weighted_effects(snowshoe_io)
    expected = sp.Matrix([
        [1, -1,  1, 1, -1],
        [1,  1, -1, 0,  1],
        [1,  1,  1, 0, -1],
        [0,  0,  1, 0, -1],
        [1,  1, -1, 0,  1]])
    assert result == expected

def test_weighted_predictions_vs_weighted_effects(snowshoe, snowshoe_io):
    result = weighted_effects(snowshoe_io)[:3, :3]
    expected = weighted_predictions_matrix(snowshoe)
    assert result == expected

def test_weighted_effects_nan_for_missing_paths(snowshoe_io_na):
    result = weighted_effects(snowshoe_io_na)
    expected = sp.Matrix([
        [     1,     -1,      1, 1,      1,     -1],
        [     1,      1,     -1, 1,      0,      1],
        [     1,      1,      1, 1,      0,     -1],
        [sp.nan, sp.nan, sp.nan, 1, sp.nan, sp.nan],
        [     0,      0,      1, 0,      0,     -1],
        [     1,      1,     -1, 1,      0,      1]])
    assert result == expected

# =============================================================================
# sign_determinacy_effects
# =============================================================================

def test_sign_determinacy_effects_average(snowshoe_io):
    result = sign_determinacy_effects(snowshoe_io, method='average')
    expected = sp.Matrix([
        [                1,                -1,  1,                 1, -1],
        [                1,                 1, -1, sp.Rational(1, 2),  1],
        [                1,                 1,  1, sp.Rational(1, 2), -1],
        [sp.Rational(1, 2), sp.Rational(1, 2),  1, sp.Rational(1, 2), -1],
        [                1,                 1, -1, sp.Rational(1, 2),  1]])
    assert result == expected

def test_sign_determinacy_effects_95_bound(snowshoe_io):
    result = sign_determinacy_effects(snowshoe_io, method='95_bound')
    expected = sp.Matrix([
        [                1,                -1,  1,                 1, -1],
        [                1,                 1, -1, sp.Rational(1, 2),  1],
        [                1,                 1,  1, sp.Rational(1, 2), -1],
        [sp.Rational(1, 2), sp.Rational(1, 2),  1, sp.Rational(1, 2), -1],
        [                1,                 1, -1, sp.Rational(1, 2),  1]])
    assert result == expected

def test_sign_determinacy_effects_vs_matrix(snowshoe, snowshoe_io):
    result = sign_determinacy_effects(snowshoe_io, method='average')[:3, :3]
    expected = sign_determinacy_matrix(snowshoe, method='average')
    assert result == expected

def test_sign_determinacy_effects_nan_for_missing_paths(snowshoe_io_na):
    result = sign_determinacy_effects(snowshoe_io_na, method='average')
    half = sp.Rational(1, 2)
    expected = sp.Matrix([
        [     1,     -1,      1,    1,      1,     -1],
        [     1,      1,     -1,    1,   half,      1],
        [     1,      1,      1,    1,   half,     -1],
        [sp.nan, sp.nan, sp.nan,    1, sp.nan, sp.nan],
        [  half,   half,      1, half,   half,     -1],
        [     1,      1,     -1,    1,   half,      1]])
    assert result == expected

# =============================================================================
# get_simulations
# =============================================================================

def test_get_simulations(snowshoe_io):
    result = set(get_simulations(snowshoe_io, n_sim=100, seed=42).keys())
    expected = {'effects', 'valid_sims', 'all_nodes', 'tmat', 'prop_stable', 'attempts', 'n_stable', 'n_valid', 'n_attempts'}
    assert result == expected

def test_get_simulations_effects_length(snowshoe_io):
    n_sim = 100
    result = len(get_simulations(snowshoe_io, n_sim=n_sim, seed=42)['effects'])
    expected = n_sim
    assert result == expected


def test_get_simulations_reproducibility(snowshoe_io):
    result_data = get_simulations(snowshoe_io, n_sim=100, seed=42)
    expected_data = get_simulations(snowshoe_io, n_sim=100, seed=42)
    result = {
        'effects': [effect.tolist() for effect in result_data['effects']],
        'valid_sims': result_data['valid_sims'],
        'all_nodes': result_data['all_nodes'],
        'tmat': result_data['tmat'].tolist(),
    }
    expected = {
        'effects': [effect.tolist() for effect in expected_data['effects']],
        'valid_sims': expected_data['valid_sims'],
        'all_nodes': expected_data['all_nodes'],
        'tmat': expected_data['tmat'].tolist(),
    }
    assert result == expected


@pytest.mark.parametrize("dist", ['uniform', 'weak', 'moderate', 'strong'])
def test_get_simulations_distributions(snowshoe_io, dist):
    result = len(get_simulations(snowshoe_io, n_sim=100, dist=dist, seed=42)['effects'])
    expected = 100
    assert result == expected

def test_get_simulations_uniform_two_oom(snowshoe_io):
    sims = get_simulations(snowshoe_io, n_sim=50, dist="uniform_two_oom", seed=42)
    assert len(sims["effects"]) == 50


def test_get_simulations_presample_applied_before_sampling(snowshoe):
    def presample(symbols):
        return {sp.Symbol('a_R,R'): 1}

    sims = get_simulations(snowshoe, n_sim=100, seed=42, presample=presample, return_samples=True)
    assert 'a_R,R' in sims['samples']
    assert np.all(sims['samples']['a_R,R'] == 1)


def test_get_simulations_presample_symbols_available(snowshoe_io):
    def presample(symbols):
        expected = {sp.Symbol('a_R,R'), sp.Symbol('b_R,Inp1')}
        assert set(symbols) >= expected
        return {sp.Symbol('a_R,R'): 1, sp.Symbol('b_R,Inp1'): 0.5}

    sims = get_simulations(snowshoe_io, n_sim=100, seed=42, presample=presample, return_samples=True)
    assert 'a_R,R' in sims['samples']
    assert 'b_R,Inp1' in sims['samples']
    assert np.all(sims['samples']['a_R,R'] == 1)
    assert np.all(sims['samples']['b_R,Inp1'] == 0.5)


def test_get_simulations_return_samples(snowshoe):
    sims = get_simulations(snowshoe, n_sim=100, seed=42, return_samples=True)
    assert 'samples' in sims
    assert all(len(v) == 100 for v in sims['samples'].values())
    assert 'a_R,R' in sims['samples']


def test_get_simulations_prop_stable(snowshoe):
    sims = get_simulations(snowshoe, n_sim=100, seed=42)
    assert sims['prop_stable'] == pytest.approx(1.0)


def test_get_simulations_with_perturb(snowshoe_io):
    state_nodes = get_nodes(snowshoe_io, 'state')
    perturb = (state_nodes[0], 1)
    effects = get_simulations(snowshoe_io, n_sim=100, seed=42, perturb=perturb)['effects']
    result = [effect.ndim for effect in effects]
    expected = [1] * len(effects)
    assert result == expected

def test_get_simulations_with_perturb_negative(snowshoe_io):
    state_nodes = get_nodes(snowshoe_io, 'state')
    perturb = (state_nodes[0], -1)
    result = 'effects' in get_simulations(snowshoe_io, n_sim=100, seed=42, perturb=perturb)
    expected = True
    assert result == expected

def test_get_simulations_with_observe(snowshoe_io):
    state_nodes = get_nodes(snowshoe_io, 'state')
    perturb = (state_nodes[0], 1)
    observe = ((state_nodes[1], 1),)
    result = 'valid_sims' in get_simulations(snowshoe_io, n_sim=100, seed=42, perturb=perturb, observe=observe)
    expected = True
    assert result == expected

def test_get_simulations_all_nodes_includes_all(snowshoe_io):
    sim_data = get_simulations(snowshoe_io, n_sim=100, seed=42)
    state_nodes = get_nodes(snowshoe_io, 'state')
    input_nodes = get_nodes(snowshoe_io, 'input')
    output_nodes = get_nodes(snowshoe_io, 'output')
    all_expected = state_nodes + input_nodes + output_nodes
    result = set(sim_data['all_nodes'])
    expected = set(all_expected)
    assert result == expected

def test_get_simulations_invalid_perturb_node(snowshoe_io):
    with pytest.raises(KeyError, match="Perturbation node 'InvalidNode' not found."):
        get_simulations(snowshoe_io, n_sim=100, perturb=('InvalidNode', 1))


def test_get_simulations_no_state_variables(io_only_graph):
    with pytest.raises(ValueError) as exc_info:
        get_simulations(io_only_graph, n_sim=100, perturb=('I', 1))
    assert "Direct input to output edge" in str(exc_info.value)


def test_get_simulations_runtime_error_max_iterations(positive_loop_graph):
    with pytest.raises(RuntimeError) as exc_info:
        get_simulations(positive_loop_graph, n_sim=100, seed=42)
    message = str(exc_info.value)
    result = message.split(' Stable')[0]
    expected = "Maximum iterations reached."
    assert result == expected



# =============================================================================
# simulation_effects
# =============================================================================

def test_simulation_effects_full_matrix(snowshoe_io):
    result = simulation_effects(snowshoe_io, n_sim=100, seed=42)
    expected = sp.Matrix([
        [  1.0,  -1.0,  1.0,  1.0, -1.0],
        [  1.0,   1.0, -1.0, 0.57,  1.0],
        [  1.0,   1.0,  1.0, 0.57, -1.0],
        [-0.54, -0.54,  1.0, 0.51, -1.0],
        [  1.0,   1.0, -1.0, 0.57,  1.0]])
    assert result == expected


def test_simulation_effects_positive_only(snowshoe_io):
    result = simulation_effects(snowshoe_io, n_sim=100, seed=42, positive_only=True)
    expected = sp.Matrix([
        [ 1.0,  0.0, 1.0,  1.0, 0.0],
        [ 1.0,  1.0, 0.0, 0.57, 1.0],
        [ 1.0,  1.0, 1.0, 0.57, 0.0],
        [0.46, 0.46, 1.0, 0.51, 0.0],
        [ 1.0,  1.0, 0.0, 0.57, 1.0]])
    assert result == expected


def test_simulation_effects_presample_full_matrix(snowshoe_rp):
    def presample(symbols):
        return {sp.Symbol('a_P,R'): 1}

    result = simulation_effects(snowshoe_rp, n_sim=100, seed=42, presample=presample, positive_only=False)
    expected = sp.Matrix([
        [  1.0,  -1.0,  1.0],
        [-0.59,   1.0, -1.0],
        [  1.0, -0.54,  1.0]])
    assert result == expected


@pytest.mark.parametrize("dist", ['uniform', 'uniform_two_oom', 'weak', 'moderate', 'strong'])
def test_simulation_effects_distributions(snowshoe_io, dist):
    expected_mats = {
        'uniform': sp.Matrix([
            [  1.0,  -1.0,  1.0,  1.0, -1.0],
            [  1.0,   1.0, -1.0, 0.57,  1.0],
            [  1.0,   1.0,  1.0, 0.57, -1.0],
            [-0.54, -0.54,  1.0, 0.51, -1.0],
            [  1.0,   1.0, -1.0, 0.57,  1.0]]),
        'uniform_two_oom': sp.Matrix([
            [  1.0,  -1.0,  1.0,  1.0, -1.0],
            [  1.0,   1.0, -1.0, 0.57,  1.0],
            [  1.0,   1.0,  1.0, 0.57, -1.0],
            [-0.54, -0.54,  1.0, 0.51, -1.0],
            [  1.0,   1.0, -1.0, 0.57,  1.0]]),
        'weak': sp.Matrix([
            [  1.0,  -1.0,  1.0,  1.0, -1.0],
            [  1.0,   1.0, -1.0, 0.58,  1.0],
            [  1.0,   1.0,  1.0, 0.58, -1.0],
            [-0.51, -0.51,  1.0, 0.57, -1.0],
            [  1.0,   1.0, -1.0, 0.58,  1.0]]),
        'moderate': sp.Matrix([
            [  1.0,  -1.0,  1.0,  1.0, -1.0],
            [  1.0,   1.0, -1.0, 0.51,  1.0],
            [  1.0,   1.0,  1.0, 0.51, -1.0],
            [-0.52, -0.52,  1.0, 0.53, -1.0],
            [  1.0,   1.0, -1.0, 0.51,  1.0]]),
        'strong': sp.Matrix([
            [  1.0,  -1.0,  1.0,   1.0, -1.0],
            [  1.0,   1.0, -1.0, -0.54,  1.0],
            [  1.0,   1.0,  1.0, -0.54, -1.0],
            [-0.56, -0.56,  1.0,  0.52, -1.0],
            [  1.0,   1.0, -1.0, -0.54,  1.0]]),
    }
    result = simulation_effects(snowshoe_io, n_sim=100, dist=dist, seed=42)
    expected = expected_mats[dist]
    assert result == expected

def test_net_effects_returns_signed_cumulative(snowshoe_io):
    result = net_effects(snowshoe_io)
    expected = cumulative_effects(snowshoe_io, form='signed')
    assert result == expected


def test_net_effects_vs_adjoint_signed(snowshoe, snowshoe_io):
    result = cumulative_effects(snowshoe_io, form='signed')[:3, :3]
    expected = adjoint_matrix(snowshoe, form='signed')
    assert result == expected

def test_simulation_effects_vs_numerical_simulations(snowshoe, snowshoe_io):
    seed = 42
    n_sim = 100
    result = simulation_effects(snowshoe_io, n_sim=n_sim, seed=seed)[:3, :3]
    expected = numerical_simulations(snowshoe, n_sim=n_sim, seed=seed)
    assert result == expected


def test_simulation_effects_nan_for_no_path(snowshoe_io_na):
    result = simulation_effects(snowshoe_io_na, n_sim=100, seed=42)
    expected = sp.Matrix([
        [   1.0,   -1.0,    1.0,  1.0,    1.0,   -1.0],
        [   1.0,    1.0,   -1.0,  1.0,   0.56,    1.0],
        [   1.0,    1.0,    1.0,  1.0,   0.56,   -1.0],
        [sp.nan, sp.nan, sp.nan,  1.0, sp.nan, sp.nan],
        [  0.52,   0.52,    1.0, 0.52,   0.50,   -1.0],
        [   1.0,    1.0,   -1.0,  1.0,   0.56,    1.0]])
    assert result == expected


def test_simulation_effects_positive_only_nan_for_no_path(snowshoe_io_na):
    result = simulation_effects(snowshoe_io_na, n_sim=100, seed=42, positive_only=True)
    expected = sp.Matrix([
        [   1.0,    0.0,    1.0,  1.0,    1.0,    0.0],
        [   1.0,    1.0,    0.0,  1.0,   0.56,    1.0],
        [   1.0,    1.0,    1.0,  1.0,   0.56,    0.0],
        [sp.nan, sp.nan, sp.nan,  1.0, sp.nan, sp.nan],
        [  0.52,   0.52,    1.0, 0.52,   0.50,    0.0],
        [   1.0,    1.0,    0.0,  1.0,   0.56,    1.0]])
    assert result == expected


# =============================================================================
# simulations_table
# =============================================================================

def test_simulations_table_no_response_nodes(minimal_error_graph):
    graph = define_input_output(minimal_error_graph)
    result = simulations_table(graph, perturb="A:+", n_sim=5, seed=42)
    expected_columns = [
        "model",
        "effect_on",
        "negative",
        "no_effect",
        "positive",
        "valid_sims",
        "stable_sims",
        "attempts",
    ]
    assert result.columns.tolist() == expected_columns
    assert result.empty


def test_simulations_table_no_valid_sims(snowshoe_io):
    result = simulations_table(snowshoe_io, perturb="P:+", observe="C:0", n_sim=50, seed=42)
    assert result["valid_sims"].eq(0).all()
    assert result["negative"].eq(0).all()
    assert result["positive"].eq(0).all()


def test_simulations_table_counts_match_structure(snowshoe_io_na):
    result = simulations_table(snowshoe_io_na, perturb="P:+", n_sim=100, seed=42)
    expected_columns = [
        "model",
        "effect_on",
        "negative",
        "no_effect",
        "positive",
        "valid_sims",
        "stable_sims",
        "attempts",
    ]
    assert result.columns.tolist() == expected_columns
    assert (result["model"] == 1).all()

    tmat = sp.matrix2numpy(absolute_effects(snowshoe_io_na)).astype(int)
    response_nodes = get_nodes(snowshoe_io_na, "state") + get_nodes(snowshoe_io_na, "output")
    perturb_nodes = get_nodes(snowshoe_io_na, "state") + get_nodes(snowshoe_io_na, "input")
    p_idx = perturb_nodes.index("P")

    for _, row in result.iterrows():
        node_idx = response_nodes.index(row["effect_on"])
        has_effect = tmat[node_idx, p_idx] != 0
        if has_effect:
            assert row["no_effect"] == 0
            assert row["negative"] + row["positive"] == row["valid_sims"]
        else:
            assert row["no_effect"] == row["valid_sims"]
            assert row["negative"] == 0
            assert row["positive"] == 0


def test_simulations_table_importable():
    from qmm.extensions import simulations_table as from_extensions
    assert simulations_table is from_extensions


# =============================================================================
# Additional coverage tests
# =============================================================================

def test_cumulative_effects_binary_form(snowshoe_io):
    result = cumulative_effects(snowshoe_io, form="binary")
    assert result.shape[0] > 0
    assert result.shape[1] > 0


def test_simulations_table_with_observe(snowshoe):
    result = simulations_table(snowshoe, perturb='R:+', observe='C:+', n_sim=100, seed=42)
    assert result is not None


# =============================================================================
# direct_effects() and table_of_direct_effects()
# =============================================================================

def test_direct_effects_net_form(snowshoe_io):
    result = direct_effects(snowshoe_io, form="net")
    assert isinstance(result, sp.MatrixBase)
    assert result.shape == (5, 5)


def test_direct_effects_absolute_form(snowshoe_io):
    result = direct_effects(snowshoe_io, form="absolute")
    assert isinstance(result, sp.MatrixBase)
    assert result.shape == (5, 5)


def test_direct_effects_positive_form(snowshoe_io):
    result = direct_effects(snowshoe_io, form="positive")
    assert isinstance(result, sp.MatrixBase)
    assert result.shape == (5, 5)


def test_direct_effects_negative_form(snowshoe_io):
    result = direct_effects(snowshoe_io, form="negative")
    assert isinstance(result, sp.MatrixBase)
    assert result.shape == (5, 5)


def test_direct_effects_invalid_form(snowshoe_io):
    with pytest.raises(ValueError, match="Invalid form"):
        direct_effects(snowshoe_io, form="invalid")


def test_table_of_direct_effects(snowshoe_io):
    result = table_of_direct_effects(snowshoe_io)
    assert isinstance(result, pd.DataFrame)
    assert result.shape == (5, 5)


# =============================================================================
# table_of_effects()
# =============================================================================

def test_table_of_effects_with_string_net_effects(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator="net_effects")
    assert isinstance(result, pd.DataFrame)


def test_table_of_effects_with_string_absolute_effects(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator="absolute_effects")
    assert isinstance(result, pd.DataFrame)


def test_table_of_effects_with_string_weighted_effects(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator="weighted_effects")
    assert isinstance(result, pd.DataFrame)


def test_table_of_effects_with_string_sign_determinacy_effects(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator="sign_determinacy_effects")
    assert isinstance(result, pd.DataFrame)


def test_table_of_effects_with_string_simulation_effects(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator="simulation_effects")
    assert isinstance(result, pd.DataFrame)


def test_table_of_effects_invalid_string_generator(snowshoe_io):
    with pytest.raises(ValueError, match="Generator must be callable"):
        table_of_effects(snowshoe_io, generator="invalid_generator")


def test_table_of_effects_with_lambda_no_name(snowshoe_io):
    result = table_of_effects(snowshoe_io, generator=lambda G: net_effects(G))
    assert isinstance(result, pd.DataFrame)
    assert result.shape == (5, 5)


def test_simulation_effects_handles_singular_matrices(snowshoe_io):
    from unittest.mock import patch, MagicMock
    original_inv = np.linalg.inv
    call_count = [0]

    def selective_inv(A):
        call_count[0] += 1
        if call_count[0] % 3 == 0:
            raise np.linalg.LinAlgError("Singular matrix")
        return original_inv(A)

    with patch('numpy.linalg.inv', side_effect=selective_inv):
        result = simulation_effects(snowshoe_io, n_sim=5, seed=42)
        assert isinstance(result, sp.Matrix)
        assert call_count[0] > 5



"""Analyse the sensitivity of system stability to direct effects within feedback cycles."""

import sympy as sp
import networkx as nx
from functools import cache
from ..core.structure import create_matrix
from ..core.stability import system_feedback, net_feedback, absolute_feedback
from ..core.helper import get_nodes, get_weight
from typing import Optional

@cache
def structural_sensitivity(G: nx.DiGraph, level: Optional[int] = None) -> sp.Matrix:
    """Calculate contribution of direct effects to stabilising and destabilising feedback.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        level: Feedback level (None for highest level)

    Returns:
        sp.Matrix: Ratio of net to total feedback terms for each direct effect

    References:
        - Hosack, G.R., Li, H.W., Rossignol, P.A. (2009). Sensitivity of system stability to model structure. Ecological Modelling 220, 1054–1062.

    Examples:
        ```python
        from qmm import structural_sensitivity, load_digraph
        structural_sensitivity(load_digraph("snowshoe_rp"))
        # Matrix([
        # [-a_C,P*a_P,C*a_R,R, a_C,P*a_P,R*a_R,C - a_C,R*a_P,P*a_R,C,                                      0],
        # [-a_C,R*a_P,P*a_R,C,                                     0, -a_C,P*a_P,C*a_R,R + a_C,P*a_P,R*a_R,C],
        # [ a_C,P*a_P,R*a_R,C,                    -a_C,P*a_P,C*a_R,R,                     -a_C,R*a_P,P*a_R,C]])
        ```
    """
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = system_feedback(G)[1:]
    if level is None:
        level = n
    if level < 1 or level > n:
        raise ValueError(f"Level must be between 1 and {n}")
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = system_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S

@cache
def net_structural_sensitivity(G: nx.DiGraph, level: Optional[int] = None) -> sp.Matrix:
    """Calculate net contribution of direct effects to system feedback.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        level: Feedback level (None for highest level)

    Returns:
        sp.Matrix: Net feedback terms containing each direct effect

    References:
        - Hosack, G.R., Li, H.W., Rossignol, P.A. (2009). Sensitivity of system stability to model structure. Ecological Modelling 220, 1054–1062.

    Examples:
        ```python
        from qmm import net_structural_sensitivity, load_digraph
        net_structural_sensitivity(load_digraph("snowshoe_rp"))
        # Matrix([
        # [-1,  0,  0],
        # [-1,  0,  0],
        # [ 1, -1, -1]])
        ```
    """
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = net_feedback(G)[1:]
    if level is None:
        level = n
    if level < 1 or level > n:
        raise ValueError(f"Level must be between 1 and {n}")
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = net_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S

@cache
def absolute_structural_sensitivity(G: nx.DiGraph, level: Optional[int] = None) -> sp.Matrix:
    """Calculate total contribution of direct effects to system feedback.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        level: Feedback level (None for highest level)

    Returns:
        sp.Matrix: Total feedback terms containing each direct effect

    References:
        - Hosack, G.R., Li, H.W., Rossignol, P.A. (2009). Sensitivity of system stability to model structure. Ecological Modelling 220, 1054–1062.

    Examples:
        ```python
        from qmm import absolute_structural_sensitivity, load_digraph
        absolute_structural_sensitivity(load_digraph("snowshoe_rp"))
        # Matrix([
        # [1, 2, 0],
        # [1, 0, 2],
        # [1, 1, 1]])
        ```
    """
    A = create_matrix(G, "signed")
    n = A.shape[0]
    fcp = absolute_feedback(G)[1:]
    if level is None:
        level = n
    if level < 1 or level > n:
        raise ValueError(f"Level must be between 1 and {n}")
    S = sp.zeros(n, n)
    nodes = get_nodes(G, "state")
    for i in range(n):
        for j in range(n):
            if A[i, j] != 0:
                sG = nx.DiGraph(G)
                sG[nodes[j]][nodes[i]]["sign"] = 0
                scp = absolute_feedback(sG)[1:]
                if level <= len(fcp) and level <= len(scp):
                    N = fcp[level - 1] - scp[level - 1]
                    S[i, j] = N
    return S

@cache
def weighted_structural_sensitivity(G: nx.DiGraph, level: Optional[int] = None) -> sp.Matrix:
    """Calculate weighted structual sensitvity for each direct effect.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        level: Feedback level (None for highest level)

    Returns:
        sp.Matrix: Weighted structural sensitivity of each direct effect

    References:
        - Hosack, G.R., Li, H.W., Rossignol, P.A. (2009). Sensitivity of system stability to model structure. Ecological Modelling 220, 1054–1062.

    Examples:
        ```python
        from qmm import weighted_structural_sensitivity, load_digraph
        weighted_structural_sensitivity(load_digraph("snowshoe_rp"))
        # Matrix([
        # [-1,   0, nan],
        # [-1, nan,   0],
        # [ 1,  -1,  -1]])
        ```
    """
    A = create_matrix(G, "signed")
    n = A.shape[0]
    if level is None:
        level = n
    if level < 1 or level > n:
        raise ValueError(f"Level must be between 1 and {n}")
    net = sp.Matrix(net_structural_sensitivity(G, level))
    absolute = sp.Matrix(absolute_structural_sensitivity(G, level))
    return get_weight(net, absolute)



"""Identify informative indicators from press perturbations."""

import pandas as pd
import numpy as np
import networkx as nx
from functools import cache
from ..core.helper import get_nodes, _parse_perturbations
from .effects import get_simulations
from typing import Union, List

@cache
def mutual_information(models: Union[nx.DiGraph, List[nx.DiGraph]], perturb: str, n_sim: int = 10000, seed: int = 42, include_null: bool = False) -> pd.DataFrame:
    """Calculate mutual information of variables for alternative models.

    Args:
        models: One or more NetworkX DiGraphs representing alternative models
        perturb: Node and sign to perturb (can be comma-separated for multiple perturbations)
        n_sim: Number of simulations
        seed: Random seed
        include_null: If True, include a null model with equal probability (1/3)
            of positive, negative, or NaN response across simulations

    Returns:
        pd.DataFrame: Mutual information for indicator selection

    References:
        - Hosack, G.R., Hayes, K.R., Dambacher, J.M. (2008). Assessing Model Structure Uncertainty Through an Analysis of System Feedback and Bayesian Networks. Ecological Applications 18, 1070–1082.
        - Melbourne-Thomas, J., Wotherspoon, S., Raymond, B., Constable, A. (2012). Comprehensive evaluation of model uncertainty in qualitative network analyses. Ecological Monographs 82, 505–519.

    Examples:
        ```python
        from qmm import mutual_information, load_digraph
        G1 = load_digraph("snowshoe_rp")
        G2 = G1.copy()
        G2.remove_edge('C', 'P')
        mutual_information((G1, G2), perturb='R:+', n_sim=1000)
        #   Node  Mutual Information
        # 0    R            0.522021
        # 1    P            0.463441
        # 2    C            0.147516
        ```
    """
    models = [models] if not isinstance(models, (list, tuple)) else list(models)
    models = [nx.DiGraph(G) if not isinstance(G, nx.DiGraph) else G for G in models]
    nodes = sorted(set(node for G in models for node in get_nodes(G, "state") + get_nodes(G, "output")))
    all_effects = []
    for G in models:
        G_modified, perturb_tuple = _parse_perturbations(G, perturb)
        sims = get_simulations(G_modified, n_sim=n_sim, seed=seed, perturb=perturb_tuple)
        response_nodes = get_nodes(G, "state") + get_nodes(G, "output")
        node_map = {node: i for i, node in enumerate(response_nodes)}
        sim_effects = []
        for effect in sims["effects"]:
            sim_effects.append([effect[node_map[n]] if n in node_map and node_map[n] < len(effect) else np.nan for n in nodes])
        all_effects.append(np.array(sim_effects))
    if include_null:
        rng = np.random.default_rng(seed)
        choices = rng.choice([1.0, -1.0, 0.0], size=(n_sim, len(nodes)))
        all_effects.append(choices)
    n_models = len(all_effects)
    mi_vals = []
    for i, node in enumerate(nodes):
        node_effects = [effects[:, i] for effects in all_effects]
        if any(np.all(np.isnan(effects)) for effects in node_effects):
            mi_vals.append(0)
            continue
        node_effects = np.concatenate(node_effects)
        labels = np.concatenate([np.full(effects.shape[0], i) for i, effects in enumerate(all_effects)])
        valid = ~np.isnan(node_effects)
        node_effects, labels = node_effects[valid], labels[valid]
        effect_signs = np.sign(node_effects) + 1
        joint, _, _ = np.histogram2d(labels, effect_signs, bins=(n_models, 3))
        joint_p = joint / joint.sum()
        l_p, e_p = joint_p.sum(axis=1), joint_p.sum(axis=0)
        mi = sum(joint_p[i, j] * np.log(joint_p[i, j] / (l_p[i] * e_p[j] + 1e-10))
                 for i in range(n_models)
                 for j in range(3)
                 if joint_p[i, j] > 0)
        mi_vals.append(max(0, mi))
    return pd.DataFrame({"Node": nodes, "Mutual Information": mi_vals}).sort_values("Mutual Information", ascending=False).reset_index(drop=True)



"""Tests for qmm.extensions.senstability module."""

import pytest
import sympy as sp

from qmm.core.helper import get_nodes
from qmm.extensions.senstability import (
    structural_sensitivity,
    net_structural_sensitivity,
    absolute_structural_sensitivity,
    weighted_structural_sensitivity,
)

# =============================================================================
# structural_sensitivity
# =============================================================================

def test_structural_sensitivity_form_symbolic_all_levels_snowshoe(snowshoe):
    result = structural_sensitivity(snowshoe)
    a_RR = sp.Symbol('a_R,R')
    a_RC = sp.Symbol('a_R,C')
    a_CR = sp.Symbol('a_C,R')
    a_CP = sp.Symbol('a_C,P')
    a_PC = sp.Symbol('a_P,C')
    a_PP = sp.Symbol('a_P,P')
    expected = sp.Matrix([
        [-a_CP*a_PC*a_RR, -a_CR*a_PP*a_RC,               0],
        [-a_CR*a_PP*a_RC,               0, -a_CP*a_PC*a_RR],
        [              0, -a_CP*a_PC*a_RR, -a_CR*a_PP*a_RC]])
    assert result == expected

def test_structural_sensitivity_form_symbolic_level_2_snowshoe(snowshoe):
    result = structural_sensitivity(snowshoe, level=2)
    a_RR = sp.Symbol('a_R,R')
    a_RC = sp.Symbol('a_R,C')
    a_CR = sp.Symbol('a_C,R')
    a_CP = sp.Symbol('a_C,P')
    a_PC = sp.Symbol('a_P,C')
    a_PP = sp.Symbol('a_P,P')
    expected = sp.Matrix([
        [-a_PP*a_RR, -a_CR*a_RC,          0],
        [-a_CR*a_RC,          0, -a_CP*a_PC],
        [         0, -a_CP*a_PC, -a_PP*a_RR]])
    assert result == expected

def test_structural_sensitivity_level_none_defaults_highest_snowshoe(snowshoe):
    n = len(get_nodes(snowshoe, 'state'))
    result = structural_sensitivity(snowshoe, level=None)
    expected = structural_sensitivity(snowshoe, level=n)
    assert result == expected

# =============================================================================
# net_structural_sensitivity
# =============================================================================

def test_net_structural_sensitivity_form_symbolic_snowshoe(snowshoe):
    result = net_structural_sensitivity(snowshoe)
    expected = sp.Matrix([
        [-1, -1,  0],
        [-1,  0, -1],
        [ 0, -1, -1]])
    assert result == expected

def test_net_structural_sensitivity_level_1_snowshoe(snowshoe):
    result = net_structural_sensitivity(snowshoe, level=1)
    expected = sp.Matrix([
        [-1, 0,  0],
        [ 0, 0,  0],
        [ 0, 0, -1]])
    assert result == expected

def test_net_structural_sensitivity_form_symbolic_chain(chain):
    result = net_structural_sensitivity(chain)
    expected = sp.Matrix([
        [-5, -3,  0,  0,  0],
        [-3, -3, -2,  0,  0],
        [ 0, -2, -4, -2,  0],
        [ 0,  0, -2, -3, -3],
        [ 0,  0,  0, -3, -5]])
    assert result == expected

# =============================================================================
# absolute_structural_sensitivity
# =============================================================================

def test_absolute_structural_sensitivity_form_symbolic_snowshoe(snowshoe):
    result = absolute_structural_sensitivity(snowshoe)
    expected = sp.Matrix([
        [1, 1, 0],
        [1, 0, 1],
        [0, 1, 1]])
    assert result == expected

def test_absolute_structural_sensitivity_level_1_snowshoe(snowshoe):
    result = absolute_structural_sensitivity(snowshoe, level=1)
    expected = sp.Matrix([
        [1, 0, 0],
        [0, 0, 0],
        [0, 0, 1]])
    assert result == expected

def test_absolute_structural_sensitivity_form_symbolic_chain(chain):
    result = absolute_structural_sensitivity(chain)
    expected = sp.Matrix([
        [5, 3, 0, 0, 0],
        [3, 3, 2, 0, 0],
        [0, 2, 4, 2, 0],
        [0, 0, 2, 3, 3],
        [0, 0, 0, 3, 5]])
    assert result == expected

# =============================================================================
# weighted_structural_sensitivity
# =============================================================================

def test_weighted_structural_sensitivity_form_symbolic_snowshoe(snowshoe):
    result = weighted_structural_sensitivity(snowshoe)
    expected = sp.Matrix([
        [    -1,     -1, sp.nan],
        [    -1, sp.nan,     -1],
        [sp.nan,     -1,     -1]])
    assert result == expected

def test_weighted_structural_sensitivity_form_symbolic_chain(chain):
    result = weighted_structural_sensitivity(chain)
    expected = sp.Matrix([
        [    -1,     -1, sp.nan, sp.nan, sp.nan],
        [    -1,     -1,     -1, sp.nan, sp.nan],
        [sp.nan,     -1,     -1,     -1, sp.nan],
        [sp.nan, sp.nan,     -1,     -1,     -1],
        [sp.nan, sp.nan, sp.nan,     -1,     -1]])
    assert result == expected


# =============================================================================
# Additional coverage tests
# =============================================================================

def test_structural_sensitivity_invalid_level_low(snowshoe):
    with pytest.raises(ValueError, match="Level must be between"):
        structural_sensitivity(snowshoe, level=0)


def test_net_structural_sensitivity_invalid_level_low(snowshoe):
    with pytest.raises(ValueError, match="Level must be between"):
        net_structural_sensitivity(snowshoe, level=0)


def test_absolute_structural_sensitivity_invalid_level_low(snowshoe):
    with pytest.raises(ValueError, match="Level must be between"):
        absolute_structural_sensitivity(snowshoe, level=0)


def test_weighted_structural_sensitivity_invalid_level_low(snowshoe):
    with pytest.raises(ValueError, match="Level must be between"):
        weighted_structural_sensitivity(snowshoe, level=0)



"""Analyse cumulative effects from perturbation scenarios with multiple-inputs and multiple-outputs."""

import numpy as np
import pandas as pd
import sympy as sp
import networkx as nx
from functools import cache
from ..core.helper import (
    get_nodes,
    get_weight,
    get_positive,
    get_negative,
    sign_determinacy,
    _random_sampler,
    _parse_perturbations,
    _parse_observations,
    get_dashed_alternatives,
    _check_direct_io_edges,
    _check_acyclic_inputs,
    _check_acyclic_outputs,
)
from ..core.structure import create_matrix
from ..core.press import (
    adjoint_matrix,
    absolute_feedback_matrix,
)
from typing import Callable, Dict, Optional, Any, Tuple, Literal, Union


def define_input_output(G: nx.DiGraph, remove_disconnected: bool = True) -> nx.DiGraph:
    """Define model components as state variables, inputs and outputs.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        remove_disconnected: Remove disconnected components

    Returns:
        nx.DiGraph: Model with input, state and output classification

    Examples:
        ```python
        from qmm import define_input_output, load_digraph
        G = load_digraph("snowshoe_io")
        G_io = define_input_output(G)
        (G_io.nodes['Inp1']['category'], G_io.nodes['Out1']['category'])
        # ('input', 'output')
        ```
    """
    if not isinstance(G, nx.DiGraph):
        raise TypeError("Input must be a networkx.DiGraph.")
    G_def = G.copy()
    if remove_disconnected:
        G_undirected = G_def.to_undirected()
        connected = list(nx.connected_components(G_undirected))
        if len(connected) > 1:
            largest = max(connected, key=len)
            nodes_to_remove = [node for system in connected if system != largest for node in system]
            G_def.remove_nodes_from(nodes_to_remove)
    nx.set_node_attributes(G_def, "state", "category")
    while True:
        reclassified = False
        for node in list(G_def.nodes()):
            if G_def.nodes[node]["category"] == "state":
                if all(G_def.nodes[pred]["category"] == "input" for pred in G_def.predecessors(node)):
                    G_def.nodes[node]["category"] = "input"
                    reclassified = True
                elif all(G_def.nodes[succ]["category"] == "output" for succ in G_def.successors(node)):
                    G_def.nodes[node]["category"] = "output"
                    reclassified = True
        if not reclassified:
            break

    _check_acyclic_inputs(G_def)
    _check_acyclic_outputs(G_def)
    _check_direct_io_edges(G_def)

    nx.freeze(G_def)
    return G_def


@cache
def direct_effects(
    G: nx.DiGraph,
    form: Literal["net", "absolute", "positive", "negative"] = "net",
) -> sp.Matrix:
    """Calculate direct effects from the signed digraph structure.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        form: Type of direct effects ('net', 'absolute', 'positive', 'negative')

    Returns:
        sp.Matrix: Direct effects for state/input columns and state/output rows
    """
    A_signed = create_matrix(G, form="signed", matrix_type="A")
    B_signed = create_matrix(G, form="signed", matrix_type="B")
    C_signed = create_matrix(G, form="signed", matrix_type="C")
    D_signed = create_matrix(G, form="signed", matrix_type="D")
    signed = sp.BlockMatrix([[A_signed, B_signed], [C_signed, D_signed]]).as_explicit()

    A_binary = create_matrix(G, form="binary", matrix_type="A")
    B_binary = create_matrix(G, form="binary", matrix_type="B")
    C_binary = create_matrix(G, form="binary", matrix_type="C")
    D_binary = create_matrix(G, form="binary", matrix_type="D")
    binary = sp.BlockMatrix([[A_binary, B_binary], [C_binary, D_binary]]).as_explicit()

    if form == "net":
        return signed
    if form == "absolute":
        return binary
    if form == "positive":
        return get_positive(signed, binary)
    if form == "negative":
        return get_negative(signed, binary)
    raise ValueError("Invalid form. Choose 'net', 'absolute', 'positive', 'negative'.")


@cache
def cumulative_effects(
    G: nx.DiGraph,
    form: Literal["symbolic", "signed", "binary"] = "symbolic",
) -> sp.Matrix:
    """Calculate cumulative effects to multiple inputs using state-space representation.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        form: Type of computation ('symbolic', 'signed', or 'binary')

    Returns:
        sp.Matrix: Cumulative effects on state variables and outputs

    Examples:
        ```python
        from qmm import load_digraph, cumulative_effects
        cumulative_effects(load_digraph("snowshoe_io"), form='symbolic')[3, 3]
        # a_C,R*a_P,C*b_R,Inp1*c_Out1,P - a_C,R*a_P,P*b_R,Inp1*c_Out1,C - a_P,C*a_R,R*b_C,Inp1*c_Out1,P + a_P,P*a_R,R*b_C,Inp1*c_Out1,C

        cumulative_effects(load_digraph("snowshoe_io"), form='signed')
        # Matrix([
        # [1, -1,  1, 2, -1],
        # [1,  1, -1, 0,  1],
        # [1,  1,  1, 0, -1],
        # [0,  0,  2, 0, -2],
        # [1,  1, -1, 0,  1]])

        cumulative_effects(load_digraph("snowshoe_io"), form='binary')
        # Matrix([
        # [1, 1, 1, 2, 1],
        # [1, 1, 1, 2, 1],
        # [1, 1, 1, 2, 1],
        # [2, 2, 2, 4, 2],
        # [1, 1, 1, 2, 1]])
        ```
    """
    B = create_matrix(G, form=form, matrix_type="B")
    C = create_matrix(G, form=form, matrix_type="C")
    D = create_matrix(G, form=form, matrix_type="D")
    if form == "symbolic":
        effects = adjoint_matrix(G, form="symbolic")
    elif form == "signed":
        effects = adjoint_matrix(G, form="signed")
    elif form == "binary":
        effects = absolute_feedback_matrix(G)
    else:
        raise ValueError("Invalid form. Choose 'symbolic', 'signed', 'binary'.")
    cemat = sp.BlockMatrix([[effects, effects * B], [C * effects, C * effects * B + D]]).as_explicit()
    if form != "symbolic":
        cemat = cemat.subs({sym: 1 for sym in cemat.free_symbols})
    return sp.expand(cemat)


def _tabulate_effects(
    G: nx.DiGraph,
    effects: Union[sp.Matrix, np.ndarray, pd.DataFrame],
) -> pd.DataFrame:
    """Format effects as a table with state/input columns and state/output rows."""
    state = get_nodes(G, "state")
    inputs = get_nodes(G, "input")
    outputs = get_nodes(G, "output")
    columns = state + inputs
    index = state + outputs

    if isinstance(effects, pd.DataFrame):
        values = effects.to_numpy()
    elif isinstance(effects, sp.Matrix):
        values = np.array(effects.tolist(), dtype=object)
    else:
        values = np.array(effects, dtype=object)

    df = pd.DataFrame(values, index=index, columns=columns)
    col_groups = ["State"] * len(state) + ["Input"] * len(inputs)
    row_groups = ["State"] * len(state) + ["Output"] * len(outputs)
    df.columns = pd.MultiIndex.from_arrays([col_groups, columns])
    df.index = pd.MultiIndex.from_arrays([row_groups, index])
    return df


@cache
def net_effects(G: nx.DiGraph) -> sp.Matrix:
    """Calculate net cumulative effects from multiple inputs.

    Args:
        G: NetworkX DiGraph representing signed digraph model

    Returns:
        sp.Matrix: Net effects on state variables and outputs

    Examples:
        ```python
        from qmm import load_digraph, net_effects
        net_effects(load_digraph("snowshoe_io"))
        # Matrix([
        # [1, -1,  1, 2, -1],
        # [1,  1, -1, 0,  1],
        # [1,  1,  1, 0, -1],
        # [0,  0,  2, 0, -2],
        # [1,  1, -1, 0,  1]])
        ```
    """
    return cumulative_effects(G, form="signed")


@cache
def absolute_effects(G: nx.DiGraph) -> sp.Matrix:
    """Calculate absolute effects from multiple inputs.

    Args:
        G: NetworkX DiGraph representing signed digraph model

    Returns:
        sp.Matrix: Total effects on state variables and outputs

    Examples:
        ```python
        from qmm import load_digraph, absolute_effects
        absolute_effects(load_digraph("snowshoe_io"))
        # Matrix([
        # [1, 1, 1, 2, 1],
        # [1, 1, 1, 2, 1],
        # [1, 1, 1, 2, 1],
        # [2, 2, 2, 4, 2],
        # [1, 1, 1, 2, 1]])
        ```
    """
    return cumulative_effects(G, form="binary")


@cache
def weighted_effects(G: nx.DiGraph) -> sp.Matrix:
    """Calculate ratio of net to total terms for predicting cumulative effects.

    Args:
        G: NetworkX DiGraph representing signed digraph model

    Returns:
        sp.Matrix: Ratio of net to total effects

    Examples:
        ```python
        from qmm import load_digraph, weighted_effects
        weighted_effects(load_digraph("snowshoe_io"))
        # Matrix([
        # [1, -1,  1, 1, -1],
        # [1,  1, -1, 0,  1],
        # [1,  1,  1, 0, -1],
        # [0,  0,  1, 0, -1],
        # [1,  1, -1, 0,  1]])
        ```
    """
    net = cumulative_effects(G, form="signed")
    absolute = cumulative_effects(G, form="binary")
    return get_weight(net, absolute)


@cache
def sign_determinacy_effects(
    G: nx.DiGraph,
    method: Literal["average", "95_bound"] = "average",
) -> sp.Matrix:
    """Calculate probability of correct sign prediction for cumulative effects.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        method: Method for computing determinacy ('average', '95_bound')

    Returns:
        sp.Matrix: Sign determinacy probabilities for effects

    Examples:
        ```python
        from qmm import load_digraph, sign_determinacy_effects
        sign_determinacy_effects(load_digraph("snowshoe_io"), method='average')
        # Matrix([
        # [  1,  -1,  1,   1, -1],
        # [  1,   1, -1, 1/2,  1],
        # [  1,   1,  1, 1/2, -1],
        # [1/2, 1/2,  1, 1/2, -1],
        # [  1,   1, -1, 1/2,  1]])

        sign_determinacy_effects(load_digraph("snowshoe_io"), method='95_bound')
        # Matrix([
        # [  1,  -1,  1,   1, -1],
        # [  1,   1, -1, 1/2,  1],
        # [  1,   1,  1, 1/2, -1],
        # [1/2, 1/2,  1, 1/2, -1],
        # [  1,   1, -1, 1/2,  1]])
        ```
    """
    net = cumulative_effects(G, form="signed")
    absolute = cumulative_effects(G, form="binary")
    weighted = get_weight(net, absolute)
    return sign_determinacy(weighted, absolute, method=method)


@cache
def get_simulations(
    G: nx.DiGraph,
    n_sim: int = 10000,
    dist: Literal["uniform", "weak", "moderate", "strong", "uniform_two_oom"] = "uniform",
    seed: int = 42,
    perturb: Optional[Tuple[str, int]] = None,
    observe: Optional[Tuple[Tuple[str, int], ...]] = None,
    presample: Optional[Callable[[Tuple[sp.Symbol, ...]], Dict[sp.Symbol, Any]]] = None,
    return_samples: bool = False,
) -> Dict[str, Any]:
    """Calculate average proportion of positive and negative effects from stable numerical simulations.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        n_sim: Number of simulations
        dist: Distribution for sampling
        seed: Random seed
        perturb: Optional perturbations (node, sign)
        observe: Optional observations (node, sign)
        presample: Optional callable that receives the tuple of free symbols and
            returns a mapping of symbol substitutions to apply before sampling.
        return_samples: If True, include dict mapping symbol names to arrays of sampled values

    Returns:
        Dict containing effects, valid_sims, all_nodes, tmat, prop_stable, attempts, and optionally samples.

    Examples:
        ```python
        from qmm import get_simulations, load_digraph
        result = get_simulations(load_digraph("snowshoe_io"), n_sim=1000, perturb=('Inp1', 1))
        result['effects'][0]
        # array([ 0.20429708, -0.79306194, -0.97014254, -0.14970721, -0.41616435])
        ```
    """

    np.random.seed(seed)

    A_sym, B_sym, C_sym, D_sym = (create_matrix(G, form="symbolic", matrix_type=m) for m in "ABCD")
    symbols_all = {s for m in (A_sym, B_sym, C_sym, D_sym) if m for s in m.free_symbols}
    symbols_all |= {sp.Symbol(f"a_{u},{v}") for u, v in G.edges}
    symbols_all = tuple(sorted(symbols_all, key=str))
    fixed_subs = {}

    if presample and symbols_all and (subs := presample(symbols_all)):
        fixed_subs = {sym: subs[sym] for sym in symbols_all if sym in subs}
        A_sym, B_sym, C_sym, D_sym = (m.subs(subs) if m else None for m in (A_sym, B_sym, C_sym, D_sym))
        symbols_sampled = tuple(sorted({s for m in (A_sym, B_sym, C_sym, D_sym) if m for s in m.free_symbols}, key=str))
    else:
        symbols_sampled = symbols_all

    state, inputs, outputs = (get_nodes(G, t) for t in ("state", "input", "output"))
    all_nodes = state + inputs + outputs
    n_x, n_u, n_y = len(state), len(inputs), len(outputs)

    response_idx = {node: i for i, node in enumerate(state + outputs)}
    perturb_nodes = state + inputs
    if perturb and perturb[0] not in perturb_nodes:
        raise KeyError(f"Perturbation node '{perturb[0]}' not found.")
    p_idx, p_sign = (perturb_nodes.index(perturb[0]), perturb[1]) if perturb else (None, 1)

    tmat = sp.matrix2numpy(absolute_effects(G)).astype(int)

    A_fn = sp.lambdify(symbols_sampled, A_sym)
    B_fn = sp.lambdify(symbols_sampled, B_sym) if n_u and B_sym else None
    C_fn = sp.lambdify(symbols_sampled, C_sym) if n_y and C_sym else None
    D_fn = sp.lambdify(symbols_sampled, D_sym) if D_sym and D_sym.shape != (0, 0) else None

    def compute_sample(values):
        A = A_fn(*values)
        if not np.all(np.real(np.linalg.eigvals(A)) < 0):
            return None
        try:
            inv_A = np.linalg.inv(-A)
        except np.linalg.LinAlgError:
            return None
        B = B_fn(*values) if B_fn else np.zeros((n_x, 0))
        C = C_fn(*values) if C_fn else np.zeros((0, n_x))
        D = D_fn(*values) if D_fn else np.zeros((n_y, n_u))
        E = np.block([[inv_A, inv_A @ B], [C @ inv_A, C @ inv_A @ B + D]]) if n_x else np.zeros((n_y, n_u))
        effect = E[:, p_idx] * p_sign if p_idx is not None else E
        return effect

    def is_valid(effect, tmat_ref):
        if not observe or tmat_ref is None:
            return True
        for node, obs in observe:
            idx = response_idx[node]
            expected = tmat_ref[idx, p_idx] != 0
            if (expected and (obs == 0 or np.sign(effect[idx]) != obs)) or (not expected and obs != 0):
                return False
        return True

    effects, valid_sims, samples = [], [], []
    attempts, max_attempts = 0, n_sim * 100
    while len(effects) < n_sim and attempts < max_attempts:
        attempts += 1
        values = _random_sampler(dist, len(symbols_sampled))
        if (effect := compute_sample(values)) is None:
            continue
        effects.append(effect)
        valid_sims.append(is_valid(effect, tmat))
        if return_samples:
            samples.append(values)

    if len(effects) < n_sim:
        raise RuntimeError(f"Maximum iterations reached. Stable proportion: {len(effects) / max_attempts:.4f}")

    result = {
        "effects": effects,
        "valid_sims": valid_sims,
        "all_nodes": all_nodes,
        "tmat": tmat,
        "prop_stable": len(effects) / attempts,
        "attempts": attempts,
        "n_stable": len(effects),
        "n_valid": int(sum(valid_sims)),
        "n_attempts": attempts,
    }
    if return_samples:
        n_samples = len(samples)
        sampled_index = {sym: i for i, sym in enumerate(symbols_sampled)}
        result_samples = {}
        for sym in symbols_all:
            if sym in sampled_index:
                idx = sampled_index[sym]
                result_samples[str(sym)] = np.array([s[idx] for s in samples])
            elif sym in fixed_subs:
                result_samples[str(sym)] = np.full(n_samples, fixed_subs[sym])
        result["samples"] = result_samples
    return result


def simulation_effects(
    G: nx.DiGraph,
    n_sim: int = 10000,
    dist: Literal["uniform", "weak", "moderate", "strong", "uniform_two_oom"] = "uniform",
    seed: int = 42,
    positive_only: bool = False,
    presample: Optional[Callable[[Tuple[sp.Symbol, ...]], Dict[sp.Symbol, Any]]] = None,
) -> sp.Matrix:
    """Performs numerical simulations of cumulative effects using random interaction strengths.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        n_sim: Number of simulations
        dist: Distribution for sampling ("uniform", "weak", "moderate", "strong")
        seed: Random seed
        positive_only: Return just the proportion of positive responses instead of sign-dominant proportions
        presample: Optional callable passed through to get_simulations

    Returns:
        SymPy Matrix containing simulation results

    Examples:
        ```python
        from qmm import load_digraph, simulation_effects
        simulation_effects(load_digraph("snowshoe_io"), n_sim=1000)
        # Matrix([
        # [   1.0,   -1.0,  1.0,   1.0, -1.0],
        # [   1.0,    1.0, -1.0, 0.511,  1.0],
        # [   1.0,    1.0,  1.0, 0.511, -1.0],
        # [-0.524, -0.524,  1.0, 0.513, -1.0],
        # [   1.0,    1.0, -1.0, 0.511,  1.0]])

        simulation_effects(load_digraph("snowshoe_io"), n_sim=1000, positive_only=True)
        # Matrix([
        # [  1.0,   0.0, 1.0,   1.0, 0.0],
        # [  1.0,   1.0, 0.0, 0.511, 1.0],
        # [  1.0,   1.0, 1.0, 0.511, 0.0],
        # [0.476, 0.476, 1.0, 0.513, 0.0],
        # [  1.0,   1.0, 0.0, 0.511, 1.0]])
        ```
    """
    sims = get_simulations(G, n_sim, dist, seed, presample=presample)
    tmat = sims["tmat"]
    n_rows, n_cols = tmat.shape

    effects = np.array(sims["effects"])
    positive = np.sum(effects > 0, axis=0)
    negative = np.sum(effects < 0, axis=0)

    smat = positive / n_sim if positive_only else np.where(
        negative > positive, -negative / n_sim, positive / n_sim
    )
    smat = [[sp.nan if not tmat[i, j] else smat[i, j] for j in range(n_cols)] for i in range(n_rows)]
    return sp.Matrix(smat)


def simulations_table(
    G: nx.DiGraph,
    perturb: str,
    observe: str = "",
    n_sim: int = 10000,
    dist: Literal["uniform", "weak", "moderate", "strong", "uniform_two_oom"] = "uniform",
    seed: int = 42,
    combinations: bool = True,
    presample: Optional[Callable[[Tuple[sp.Symbol, ...]], Dict[sp.Symbol, Any]]] = None,
) -> pd.DataFrame:
    """Summarise simulation effects across model variants for each response node.

    Args:
        G: NetworkX DiGraph representing signed digraph model
        perturb: Node and sign to perturb (perturbation string)
        observe: Observation string (node:sign, comma-separated allowed) to filter simulations
        n_sim: Number of simulations
        dist: Distribution for sampling
        seed: Random seed
        combinations: If True, evaluate every combination of dashed edges
        presample: Optional callable passed through to get_simulations

    Returns:
        pd.DataFrame: Table of counts for negative, no effect, and positive responses

    Examples:
        ```python
        from qmm import load_digraph, simulations_table
        simulations_table(load_digraph("snowshoe_io"), perturb='Inp1:+', n_sim=1000)
        #    model effect_on  negative  no_effect  positive  valid_sims  stable_sims  attempts
        # 0      1         R         0          0      1000        1000         1000      1000
        # 1      1         C       489          0       511        1000         1000      1000
        # 2      1         P       489          0       511        1000         1000      1000
        # 3      1      Out1       487          0       513        1000         1000      1000
        # 4      1      Out2       489          0       511        1000         1000      1000
        ```
    """
    variants = get_dashed_alternatives(G, combinations=combinations)
    observations = _parse_observations(observe) if observe else None
    rows = []

    for model_idx, g in enumerate(variants, start=1):
        graph, pert = _parse_perturbations(g, perturb)
        sims = get_simulations(
            graph,
            n_sim=n_sim,
            dist=dist,
            seed=seed,
            perturb=pert,
            observe=observations,
            presample=presample,
        )

        response_nodes = get_nodes(g, "state") + get_nodes(g, "output")
        if not response_nodes:
            continue
        node_count = len(response_nodes)
        p_idx = sims["all_nodes"].index(pert[0])
        tmat = sims["tmat"][:node_count, :]

        valid_effects = [effect[:node_count] for effect, valid in zip(sims["effects"], sims["valid_sims"]) if valid]
        valid_count = len(valid_effects)
        if valid_count:
            effect_arr = np.array(valid_effects)
            negative = np.sum(effect_arr < 0, axis=0).astype(int)
            positive = np.sum(effect_arr > 0, axis=0).astype(int)
        else:
            negative = np.zeros(node_count, dtype=int)
            positive = np.zeros(node_count, dtype=int)
        has_effect = tmat[:, p_idx] != 0
        no_effect = np.where(has_effect, 0, valid_count).astype(int)
        negative = np.where(has_effect, negative, 0).astype(int)
        positive = np.where(has_effect, positive, 0).astype(int)

        for i, node in enumerate(response_nodes):
            row = {
                "model": model_idx,
                "effect_on": node,
                "negative": int(negative[i]),
                "no_effect": int(no_effect[i]),
                "positive": int(positive[i]),
                "valid_sims": int(valid_count),
                "stable_sims": int(sims["n_stable"]),
                "attempts": int(sims["attempts"]),
            }
            rows.append(row)

    cols = ["model", "effect_on", "negative", "no_effect", "positive", "valid_sims", "stable_sims", "attempts"]
    return pd.DataFrame(rows, columns=cols)


def table_of_direct_effects(
    G: nx.DiGraph,
    form: Literal["net", "absolute", "positive", "negative"] = "net",
) -> pd.DataFrame:
    """Create a table of direct effects with state/input columns and state/output rows."""
    return _tabulate_effects(G, direct_effects(G, form=form))


def table_of_effects(
    G: nx.DiGraph,
    generator: Union[
        Callable[..., Union[sp.Matrix, np.ndarray, pd.DataFrame]],
        Literal[
            "net_effects",
            "absolute_effects",
            "weighted_effects",
            "sign_determinacy_effects",
            "simulation_effects",
        ],
    ] = net_effects,
) -> pd.DataFrame:
    """Create a table of effects with state/input columns and state/output rows."""
    generator_name = None
    if isinstance(generator, str):
        generator_name = generator
        generator = {
            "net_effects": net_effects,
            "absolute_effects": absolute_effects,
            "weighted_effects": weighted_effects,
            "sign_determinacy_effects": sign_determinacy_effects,
            "simulation_effects": simulation_effects,
        }.get(generator)
    if not callable(generator):
        raise ValueError(f"Generator must be callable, got: {type(generator)}")
    if generator_name is None:
        generator_name = getattr(generator, "__name__", None)

    effects = generator(G)
    if generator_name in {"weighted_effects", "sign_determinacy_effects", "numerical_simulations"}:
        if isinstance(effects, sp.MatrixBase):
            effects = effects.evalf(2)
    return _tabulate_effects(G, effects)



"""Tests for qmm.extensions.validation module."""

import pytest
import pandas as pd
import numpy as np
from qmm.extensions.validation import (
    marginal_likelihood,
    model_validation,
    posterior_predictions,
    diagnose_observations,
    bayes_factors,
)

# =============================================================================
# marginal_likelihood
# =============================================================================

def test_marginal_likelihood(mesocosm):
    result = marginal_likelihood(mesocosm, perturb='P:+', observe='A1:+', seed=42)
    expected = 0.3461
    assert np.allclose(result, expected)

def test_marginal_likelihood_three_observations(mesocosm):
    result = marginal_likelihood(mesocosm, perturb='P:+', n_sim=100, observe='AP:+, C2:+, H2:+', dist='uniform', seed=42)
    expected = 0.57
    assert np.allclose(result, expected)

def test_marginal_likelihood_four_observations(mesocosm):
    result = marginal_likelihood(mesocosm, perturb='P:+', n_sim=100, observe='A2:+, AP:+, C2:+, H2:+', dist='uniform', seed=42)
    expected = 0.47
    assert np.allclose(result, expected)

def test_marginal_likelihood_reproducibility(mesocosm):
    result = marginal_likelihood(mesocosm, perturb='P:+', observe='A1:+', seed=42)
    expected = marginal_likelihood(mesocosm, perturb='P:+', observe='A1:+', seed=42)
    assert np.allclose(result, expected)

def test_marginal_likelihood_invalid_perturbation(mesocosm):
    with pytest.raises(ValueError) as exc_info:
        marginal_likelihood(mesocosm, perturb='Invalid:+', observe='A1:+')
    assert "Unknown perturbation node" in str(exc_info.value)

def test_marginal_likelihood_zero_observation_dashed_edge(snowshoe_io_na):
    G = snowshoe_io_na.copy()
    G.add_edge('R', 'N', sign=1, dashes=True)
    result = marginal_likelihood(G, perturb='P:+', observe='N:0', n_sim=100, dist='uniform', seed=42)
    expected = 0.0
    assert np.allclose(result, expected)

def test_marginal_likelihood_zero_observation_no_edge(snowshoe_io_na):
    result = marginal_likelihood(snowshoe_io_na, perturb='P:+', observe='N:0', n_sim=100, dist='uniform', seed=42)
    expected = 1.0
    assert np.allclose(result, expected)

# =============================================================================
# model_validation
# =============================================================================

def test_model_validation_alternative_structure(snowshoe_dashed):
    df = model_validation(snowshoe_dashed, perturb='C:+', observe='P:-', n_sim=100, seed=42, combinations=True)
    expected_data = {
        'Marginal likelihood': ['0.520', '0.470', '0.430', '0.430', '0.000', '0.000', '0.000', '0.000'],
        'R $\\rightarrow$ P': ['\u2713', '\u2713', '\u2713', '\u2713', '', '', '', ''],
        'C $\\multimap$ C': ['\u2713', '', '', '\u2713', '', '\u2713', '', '\u2713'],
        'P $\\multimap$ R': ['\u2713', '\u2713', '', '', '', '', '\u2713', '\u2713']
    }
    result = df.to_dict('list')
    expected = expected_data
    assert result == expected

def test_model_validation_combinations_false(snowshoe_dashed):
    df = model_validation(snowshoe_dashed, perturb='C:+', observe='P:-', n_sim=100, seed=42, combinations=False)
    expected_data = {
        'Marginal likelihood': ['0.520', '0.000'],
        'R $\\rightarrow$ P': ['\u2713', ''],
        'C $\\multimap$ C': ['\u2713', ''],
        'P $\\multimap$ R': ['\u2713', '']
    }
    result = df.to_dict('list')
    expected = expected_data
    assert result == expected

def test_model_validation_no_dashed_edges(snowshoe):
    df = model_validation(snowshoe, perturb='R:+', observe='C:+')
    result = (len(df), 'Marginal likelihood' in df.columns)
    expected = (1, True)
    assert result == expected



# =============================================================================
# posterior_predictions
# =============================================================================

def test_posterior_predictions(mesocosm):
    result = posterior_predictions(mesocosm, perturb='P:+', observe='A1:+', seed=42)
    expected = [
        1.0,
        1.0,
        -0.551863623230280,
        1.0,
        0.983819705287489,
        1.0,
        0.587113550996822,
        0.983819705287489,
    ]
    assert np.allclose(np.array(result.tolist(), dtype=float).ravel(), expected)

def test_posterior_predictions_complex_observations(mesocosm):
    result = posterior_predictions(mesocosm, perturb='P:+', n_sim=100, observe='AP:+, C2:+, H2:+', dist='uniform', seed=42)
    expected = [
        1.0,
        -0.508771929824561,
        0.824561403508772,
        1.0,
        1.0,
        1.0,
        -0.701754385964912,
        1.0,
    ]
    assert np.allclose(np.array(result.tolist(), dtype=float).ravel(), expected)

def test_posterior_predictions_positive_only(mesocosm):
    result = posterior_predictions(mesocosm, perturb='P:+', n_sim=100, observe='A2:+, AP:+, C2:+, H2:+', dist='uniform', seed=42, positive_only=True)
    expected = [
        1.0,
        0.382978723404255,
        1.0,
        1.0,
        1.0,
        1.0,
        0.191489361702128,
        1.0,
    ]
    assert np.allclose(np.array(result.tolist(), dtype=float).ravel(), expected)


def test_posterior_predictions_no_matching_simulations(mesocosm):
    result = posterior_predictions(mesocosm, perturb='P:+', observe='A1:-, A2:-, AP:-, H1:-, H2:-, C1:-, C2:-', n_sim=100, seed=42)
    expected = (8, 1)
    assert result.shape == expected

# =============================================================================
# diagnose_observations
# =============================================================================

def test_diagnose_observations_input_only(snowshoe_io):
    df = diagnose_observations(snowshoe_io, observe='R:+, C:+', n_sim=100, perturb_nodes='input', seed=42)
    expected_data = {
        'Input': ['Inp1', 'Inp1', 'Inp2', 'Inp2'],
        'Sign': ['+', '-', '+', '-'],
        'Marginal likelihood': [0.57, 0.0, 0.0, 0.0]
    }
    result = df.to_dict('list')
    expected = expected_data
    assert result == expected


def test_diagnose_observations_state_only(snowshoe_io):
    df = diagnose_observations(snowshoe_io, observe='R:+', n_sim=100, perturb_nodes='state', seed=42)
    result = set(df['Input'].unique())
    expected = {'R', 'C', 'P'}
    assert result == expected


def test_diagnose_observations_comma_separated_nodes(snowshoe_io):
    df = diagnose_observations(snowshoe_io, observe='R:+', n_sim=100, perturb_nodes='R, C', seed=42)
    result = set(df['Input'].unique())
    expected = {'R', 'C'}
    assert result == expected


def test_diagnose_observations_default_nodes(snowshoe_io):
    df = diagnose_observations(snowshoe_io, observe='R:+', n_sim=100, seed=42)
    result = set(df['Input'].unique())
    expected = {'R', 'C', 'P', 'Inp1', 'Inp2'}
    assert result == expected


def test_diagnose_observations_with_errors(snowshoe_io):
    df = diagnose_observations(snowshoe_io, observe='InvalidNode:+', n_sim=100, perturb_nodes='input', seed=42)
    result = isinstance(df, pd.DataFrame)
    expected = True
    assert result == expected


def test_diagnose_observations_all_errors_empty_df(minimal_error_graph):
    df = diagnose_observations(minimal_error_graph, observe='A:+', n_sim=10, perturb_nodes='A', seed=42)
    result = isinstance(df, pd.DataFrame)
    expected = True
    assert result == expected


# =============================================================================
# bayes_factors
# =============================================================================

def test_bayes_factors(mesocosm_alt_models):
    G, G_alt = mesocosm_alt_models
    df = bayes_factors((G, G_alt), perturb='P:+', n_sim=100, observe='AP:+, C2:+, H2:+', dist='uniform', seed=42)
    result = (
        df['Model comparison'].tolist(),
        np.allclose(df['Likelihood 1'].to_numpy(), [0.57]),
        np.allclose(df['Likelihood 2'].to_numpy(), [0.73]),
        np.allclose(df['Bayes factor'].to_numpy(), [0.781], atol=0.01)
    )
    expected = (['Model A/Model B'], True, True, True)
    assert result == expected

def test_bayes_factors_more_observations(mesocosm_alt_models):
    G, G_alt = mesocosm_alt_models
    df = bayes_factors((G, G_alt), perturb='P:+', n_sim=100, observe='A2:+, AP:+, C2:+, H2:+', dist='uniform', seed=42)
    result = (
        df['Model comparison'].tolist(),
        np.allclose(df['Likelihood 1'].to_numpy(), [0.47]),
        np.allclose(df['Likelihood 2'].to_numpy(), [0.36]),
        np.allclose(df['Bayes factor'].to_numpy(), [1.306], atol=0.01)
    )
    expected = (['Model A/Model B'], True, True, True)
    assert result == expected

def test_bayes_factors_custom_names(bayes_models):
    df = bayes_factors(bayes_models, perturb='R:+', observe='P:+', names=['ModA', 'ModB'])
    result = 'ModA/ModB' in df['Model comparison'].values
    expected = True
    assert result == expected
